/*
*	JTUGens
*	by Jan Truetzschler v. Falkenstein 2002/2003
*	jt@kandos.de
* 
*	the most recent version you can find at
*	http://www.sampleAndHold.org/supercollider/
*	includes modified code by Dan Trueman (munger)
*	and James McCartney (RecPBufDelayL),
*	the ImpulseDrop UGen was made 
*	in Collaboration with Alberto de Campo, Zlatko B., Tom Tlalim
*
*	
*	This program is free software; For License look in License.rtf
*	which you should have received with this file.
*	If not wright to jt@kandos.de
*
* 	This program is distributed in the hope that it will be useful,
*  	but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
*	
*/


#include "SC_PlugIn.h"
#include "SC_InterfaceTable.h"
//#include "JtUGens.h"

static InterfaceTable *ft;

#define NUMVOICES 100						//max number of voices
#define MINSPEED .001 						//minimum speed through buffer on playback
#define PITCHTABLESIZE 100 					//max number of transpositions for the "scale" message
#define MINSIZE 1024.0							// twice ENVSIZE. minimum grainsize in samples



///////////////////////////////////////////////////////////////////////////

struct RecPBufDelayL : public Unit
{
	float m_fbufnum;
	SndBuf *m_buf;
	float m_dsamp, m_x1, m_y1;
	float m_delaytime;
	long m_iwrphase;
	long m_irdphase;
	long m_numoutput;
	long m_recordOn;
};

struct voiceInfo {
	
	double gvoiceSpeed, gvoiceCurrent;
	float gvoiceLPan, gvoiceRPan, gvoiceGain, gvoiceOneOverRamp, gvoiceRamp;
	long gvoiceOn, gvoiceSize, gvoiceDone, gvoiceDirection;
};
typedef voiceInfo voiceInfo;
struct Munger : public Unit
{
	voiceInfo * voices;
	float m_fbufnum;
	SndBuf *m_buf; //recordBuf;
	float m_delaytime;
	float twelfth; //1/12
	double semitone;
	//float position; //playback position (0-1) (if ==-1, then random, which is default)
	
	float buflen;
	float maxsize, minsize;
	float twothirdBufsize, onethirdBufsize;
	float initbuflen;
	long maxvoices;
    float rampLength; //for simple linear ramp
	double * pitchTable; 	//table of pitch values to draw from
    float gimme;    
    long time; 
    float srate;
    float srate_ms;
	float minspeed;
    //window stuff Hanning not supported yet
	/*
	short doHanning;
	float winTime[NUMVOICES], winRate[NUMVOICES];
    float winTable[WINLENGTH];
	*/
};


struct MungerMono : public Unit
{
	voiceInfo * voices;
	float m_fbufnum;
	SndBuf *m_buf; //recordBuf;
	float m_delaytime;
	float twelfth; //1/12
	double semitone;
	//float position; //playback position (0-1) (if ==-1, then random, which is default)
	
	float buflen;
	float maxsize, minsize;
	float twothirdBufsize, onethirdBufsize;
	float initbuflen;
	long maxvoices;
    float rampLength; //for simple linear ramp
	double * pitchTable; 	//table of pitch values to draw from
    float gimme;    
    long time; 
    float srate;
    float srate_ms;
    //window stuff Hanning not supported yet
	/*
	short doHanning;
	float winTime[NUMVOICES], winRate[NUMVOICES];
    float winTable[WINLENGTH];
	*/
};

struct ImpulseDrop : public Unit
{
	double mPhase;
	float mFreqMul;
	float mProb;
};
struct PulseCount2 : public Unit
{
	float mLevel;
	float m_prevtrig, m_prevreset;
};
struct PulseCount3 : public Unit
{
	float mLevel;
	float m_prevtrig, m_prevreset;
};
struct TIUrn : public Unit
{
	float m_trig, m_value;
	//float* lastValues;
	float lastValue;
	long index;
};



extern "C"
{
	void load(InterfaceTable *inTable);
	//void setsysbar();
//Munger:
	void Munger_Ctor(Munger *unit);
	void Munger_Dtor(Munger *unit);
	void Munger_next(Munger *unit, int inNumSamples);
//monoMunger
	void MungerMono_Ctor(MungerMono *unit);
	void MungerMono_Dtor(MungerMono *unit);
	void MungerMono_next(MungerMono *unit, int inNumSamples);	



//	RecPBufDelayL:
	void RecPBufDelayL_Ctor(RecPBufDelayL *unit);
	void RecPBufDelayL_next(RecPBufDelayL *unit, int inNumSamples);
	void RecPBufDelayL_next_z(RecPBufDelayL *unit, int inNumSamples);

	void ImpulseDrop_next_a(ImpulseDrop *unit, int inNumSamples);
	void ImpulseDrop_next_k(ImpulseDrop *unit, int inNumSamples);
	void ImpulseDrop_Ctor(ImpulseDrop* unit);
	
	void PulseCount2_Ctor(PulseCount2 *unit);
	void PulseCount2_next(PulseCount2 *unit, int inNumSamples);

	void PulseCount3_Ctor(PulseCount3 *unit);
	void PulseCount3_next(PulseCount3 *unit, int inNumSamples);
	
	void TIUrn_Ctor(TIUrn *unit);
	void TIUrn_next(TIUrn *unit, int inNumSamples);
	void TIUrn_Dtor(TIUrn *unit);
	}


///////////////////////////////////////////////////////////////////////////
#define RGET \
	RGen& rgen = *unit->mParent->mRGen; \
	uint32 s1 = rgen.s1; \
	uint32 s2 = rgen.s2; \
	uint32 s3 = rgen.s3; 

#define RPUT \
	rgen.s1 = s1; \
	rgen.s2 = s2; \
	rgen.s3 = s3;

#define GET_BUF \
	float fbufnum  = ZIN0(0); \
	if (fbufnum != unit->m_fbufnum) { \
		int bufnum = (int)fbufnum; \
		World *world = unit->mWorld; \
		if (bufnum < 0 || bufnum >= world->mNumSndBufs) bufnum = 0; \
		unit->m_fbufnum = fbufnum; \
		unit->m_buf = world->mSndBufs + bufnum; \
	} \
	SndBuf *buf = unit->m_buf; \
	float *bufData __attribute__((__unused__)) = buf->data; \
	int bufChannels __attribute__((__unused__)) = buf->channels; \
	int bufSamples __attribute__((__unused__)) = buf->samples; \
	int bufFrames = buf->frames; \
	int mask __attribute__((__unused__)) = buf->mask; \
	int guardFrame __attribute__((__unused__)) = bufFrames - 2; 

#define CHECK_BUF \
	if (!bufData) { \
		ClearUnitOutputs(unit, inNumSamples); \
		return; \
	} 
#define SETUP_OUT \
	if (unit->mNumOutputs != bufChannels) { \
		ClearUnitOutputs(unit, inNumSamples); \
		return; \
	} \
	float *out[16]; \
	for (int i=0; i<bufChannels; ++i) out[i] = ZOUT(i); 

#define SETUP_IN(offset) \
	if ((unit->mNumInputs - offset) != bufChannels) { \
		ClearUnitOutputs(unit, inNumSamples); \
		return; \
	} \
	float *in[16]; \
	for (int i=0; i<bufChannels; ++i) in[i] = ZIN(i+offset); 
	
	
#define BufCalcDelay(delaytime) (sc_clip(delaytime * SAMPLERATE, 1.f, (float)bufSamples))

#define NUMVOICES 100						//max number of voices
#define MINSPEED .001 						//minimum speed through buffer on playback
#define PITCHTABLESIZE 100 					//max number of transpositions for the "scale" message

void Munger_Dtor(Munger* unit)
{
	RTFree(unit->mWorld, unit->voices); 
}

/////////Munger//////////
void Munger_Ctor(Munger *unit)
{
	int i;
	if (INRATE(1) == calc_FullRate) {
		SETCALC(Munger_next);
	} else {
		SETCALC(Munger_next);
	}
	unit->initbuflen = ZIN0(1); //was seconds is now samples
	//Print("unit->m_delaytime %g\n", unit->m_delaytime);
	unit->m_fbufnum = -1.;
	
	unit->srate = SAMPLERATE;
    unit->srate_ms = unit->srate/1000.;
    //used bufferlength
    unit->buflen = unit->initbuflen;
    unit->maxsize = unit->buflen / 3.;
    unit->onethirdBufsize = unit->maxsize;
    unit->minsize = MINSIZE;
	unit->minspeed = MINSPEED;
	unit->semitone = sc_pow(2., 1./12.);
    unit->time = 0;	
    unit->gimme = 0.f;	
    unit->maxvoices = 100;
	//unit->doHanning = 0;
	
	int size1 = NUMVOICES * sizeof(voiceInfo);
	int size2 = PITCHTABLESIZE * sizeof(double);

	char* space = (char*)RTAlloc(unit->mWorld, size1 + size2);
	unit->voices = (voiceInfo*)space;
	space += size1;
	unit->pitchTable = (double*)space;

    for(i=0; i<NUMVOICES; i++) {
    	(unit->voices + i)->gvoiceSize= 100;			
    	(unit->voices + i)->gvoiceSpeed = 1.;			
    	(unit->voices + i)->gvoiceCurrent = 0.;
    	(unit->voices + i)->gvoiceDirection = 1;		
    	(unit->voices + i)->gvoiceOn = 0;			
    	(unit->voices + i)->gvoiceDone = 0;			
    	(unit->voices + i)->gvoiceRPan = .5;
    	(unit->voices + i)->gvoiceLPan = .5;
    	(unit->voices + i)->gvoiceGain = 1.;
		(unit->voices + i)->gvoiceOneOverRamp = 0.;
		(unit->voices + i)->gvoiceRamp = 0.;
    }
    //tempered scale
    
   //for(i=0; i<PITCHTABLESIZE-1; i += 2) {
   for(i=0; i<99; i += 2) {
		unit->pitchTable[i] = (double) 0.5*i;
		unit->pitchTable[i+1] =  (double)  -0.5*i;
	}
    unit->rampLength = 512.;
	//Munger_next(unit,1);
}




void Munger_next(Munger *unit, int inNumSamples)
{
	
	float grate, grate_var, glen, glen_var, ggain, ggain_var, position, delaytime, gpan_var, rampLength;
	//float grateRes = 0.f;
	int32 ambi;
	//int recordOn;
	float outsampL, outsampR;
	int newvoice;
	float minspeed = unit->minspeed;
	float *out = 	ZOUT(0);
	float *outR = 	ZOUT(1);  
	delaytime = 	ZIN0(1); //is fixed not needed 
	grate 		= 	ZIN0(2);
	grate_var 	= 	ZIN0(3);
	glen 		= 	ZIN0(4);
	glen_var 	= 	ZIN0(5);
	double gpitch = (double) ZIN0(6);
	double gpitch_var = (double) ZIN0(7);
	ggain		= 	ZIN0(8); 
	ggain_var	= 	ZIN0(9); 
		
	ambi = (int32)	ZIN0(10);
	position = 		ZIN0(11);
	int32 numVoices = (int32) ZIN0(12);
	gpan_var	= 	ZIN0(13); 
	int32 smoothPitch = (int32) ZIN0(14);
	rampLength = 	ZIN0(15);
	//unit->rampLength = rampLength;
	//unit->gpan_spread = gpan_var;
	
	GET_BUF
	CHECK_BUF
	
	voiceInfo * voiceinfos =  unit->voices;
	
	float gimme = unit->gimme;
	int32 time = unit->time;
	//other munger:
	long buflen = (long) unit->buflen;
	//float dsamp = unit->buflen;
	double semitone = unit->semitone;
	
	float newsize;
	int32 pitchChoice;
	double pitchExponent;
	
	double alpha, om_alpha, output, where;
	long first;
	long done, tail;
	
	
	
	double newPosition;	
	float value = 0.f;
	
	int foundOne;
	newvoice = 0;
	outsampR = 0.f;
	outsampL = 0.f;
	
	double * pitchTable = unit->pitchTable;

	float srate_ms = unit->srate_ms;
	float onethirdBufsize = unit->onethirdBufsize;
	float maxsize = unit->maxsize;
	float minsize = unit->minsize;
	RGET
/*
	float* dlybuf1 = bufData - ZOFF;
	float* dlywr   = dlybuf1 + (iwrphase & mask);
	float* dlyN    = dlybuf1 + bufSamples;	
*/
	
for (int32 k=0; k<inNumSamples; ++k) { 

///--------------
		if(time++ >= (long)gimme) 
		{
			time = 0;
///--------------- find voice:
			int i = 0; 
			foundOne = -1;
			while(foundOne < 0 && i < numVoices) { //was voices
				if (!(voiceinfos +i)->gvoiceOn) foundOne = i;
				i++;
			}
			newvoice = foundOne;
//-------
		if(newvoice >= 0) {
	
////// new Size---------------
	
	//set grain pitch
			if(smoothPitch == 1) gpitch = gpitch +  ( (double) frand2(s1,s2,s3)) * gpitch_var;
			else {
				pitchChoice = (int32) ( ((double) frand(s1,s2,s3) ) * gpitch_var * 10.0);
				pitchChoice = (int32) sc_clip(pitchChoice, 0, 99);
				pitchExponent = pitchTable[pitchChoice];
				gpitch =  gpitch * sc_pow(semitone, pitchExponent);
				
			}
			(voiceinfos +newvoice)->gvoiceSpeed =  gpitch;
			if((voiceinfos +newvoice)->gvoiceSpeed < (double) minspeed) (voiceinfos +newvoice)->gvoiceSpeed = minspeed;
			
			newsize = srate_ms * (glen + frand2(s1,s2,s3) * glen_var);
			
			if(newsize > maxsize) newsize = maxsize;
			if(newsize * gpitch > maxsize) newsize = maxsize/gpitch;
			if(newsize < minsize) newsize = minsize;
			(voiceinfos +newvoice)->gvoiceSize =  (long) newsize;
	//end new size
	
//newDirection------------------
			if(ambi == 0) {
				(voiceinfos +newvoice)->gvoiceDirection = (int) frand2(s1,s2,s3);
				if((voiceinfos +newvoice)->gvoiceDirection < 0) 
					(voiceinfos +newvoice)->gvoiceDirection = -1;
				else (voiceinfos +newvoice)->gvoiceDirection = 1;
			} else {
				if(ambi < 0) (voiceinfos +newvoice)->gvoiceDirection= -1; //was ambi==-1
				else (voiceinfos +newvoice)->gvoiceDirection = 1;
			}
//end newDirection-----------------
			float nPan = frand2(s1,s2,s3) * gpan_var + 0.5f;
			float rPan =  1.f - nPan;
			(voiceinfos +newvoice)->gvoiceLPan 	= nPan;
			(voiceinfos +newvoice)->gvoiceRPan	= rPan;
			(voiceinfos +newvoice)->gvoiceOn 	= 1;
			(voiceinfos +newvoice)->gvoiceDone	= 0;
			(voiceinfos +newvoice)->gvoiceGain = ggain + frand2(s1,s2,s3) * ggain_var;
//envelope
		float gvoiceRamp = (voiceinfos +newvoice)->gvoiceRamp;
			if(((voiceinfos +newvoice)->gvoiceSize) < 2.*rampLength) {
				gvoiceRamp = .5 * (voiceinfos + newvoice)->gvoiceSize;
				if(gvoiceRamp <= 0.) gvoiceRamp = 1.;
			(voiceinfos +newvoice)->gvoiceOneOverRamp = 1./gvoiceRamp;
			} else {
				gvoiceRamp = rampLength;
				if(gvoiceRamp <= 0.) gvoiceRamp = 1.;
				(voiceinfos +newvoice)->gvoiceOneOverRamp = 1./gvoiceRamp;
			}
			
	 (voiceinfos +newvoice)->gvoiceRamp = gvoiceRamp;
/*** set start point; tricky, cause of moving buffer, variable playback rates, backwards/forwards, etc.... ***/	
	// 3. random positioning and fixed buffer	
			 if (position == -1.) {
				if((voiceinfos +newvoice)->gvoiceDirection == 1) {//going forward			
					newPosition = (double) onethirdBufsize - frand(s1,s2,s3) * onethirdBufsize;
				}	
				else //going backwards
					newPosition = (double) frand(s1,s2,s3) * onethirdBufsize;
				}
	// 4. fixed positioning and fixed buffer	
			else if (position >= 0.) {
				if((voiceinfos +newvoice)->gvoiceDirection == 1) {//going forward			
					newPosition = (double) onethirdBufsize - position * onethirdBufsize;
				}	
			else //going backwards
				newPosition = (double) position * onethirdBufsize;
			}
			(voiceinfos +newvoice)->gvoiceCurrent = newPosition;
///////end new setup---------------------------------------------------------
		}
		gimme = srate_ms * (grate  + frand(s1,s2,s3) * grate_var);//grate is actually time-distance between grains
		}
		
//mix 'em, pan 'em -------------------------------------
	//i = voices;
	//while(i--){
	for(int j = 0; j<numVoices; j++){
		if((voiceinfos + j)->gvoiceOn) {
			where = (voiceinfos + j)->gvoiceCurrent;

//getSamp -/lin interpol------------------------------------------------
			while(where < 0.) where += buflen;
			while(where > guardFrame) where -= buflen;

			first = (long)where;
			alpha = (double) where - first;
			om_alpha = 1. - alpha;		
			output = bufData[first++ & mask] * om_alpha;
			if(first <  buflen) {
				output += bufData[first & mask] * alpha;
			}
			else {
				output +=  bufData[0 & mask] * alpha;
			}
			outsampL = (float) output;
	
	
//envelope---------------------------------------------------

			done = (voiceinfos + j)->gvoiceDone;
			tail = (voiceinfos + j)->gvoiceSize - done;
	
			if(done < (voiceinfos + j)->gvoiceRamp) outsampL *= (done * (voiceinfos + j)->gvoiceOneOverRamp);
			else if(tail < (voiceinfos + j)->gvoiceRamp) outsampL *= (tail * (voiceinfos + j)->gvoiceOneOverRamp);
	
//end env---------------------
			outsampL = outsampL * (voiceinfos + j)->gvoiceGain;
		//pan it, dumb linear
			value += outsampL * (voiceinfos + j)->gvoiceLPan;
			outsampR += outsampL * (voiceinfos + j)->gvoiceRPan;
		//see if grain is done after jumping to next sample point
			(voiceinfos + j)->gvoiceCurrent += (double) (voiceinfos + j)->gvoiceDirection * (voiceinfos + j)->gvoiceSpeed;
			if(++(voiceinfos + j)->gvoiceDone >= (voiceinfos + j)->gvoiceSize) (voiceinfos + j)->gvoiceOn = 0;
				}
			}//end mix em			
				ZXP(out) = value;
				ZXP(outR) = outsampR;

	value = 0.f;
	//just make sure to reset ereything
	newvoice = 0;
	outsampR = 0.f;
	outsampL = 0.f;
		}
	RPUT
	//write back :	
	unit->gimme = gimme;
	unit->time = time;
	unit->voices = voiceinfos;

}

void MungerMono_Dtor(MungerMono* unit)
{
	RTFree(unit->mWorld, unit->voices); 
}

/////////Munger//////////
void MungerMono_Ctor(MungerMono *unit)
{
	int i;
	if (INRATE(1) == calc_FullRate) {
		SETCALC(Munger_next);
	} else {
		SETCALC(Munger_next);
	}
	unit->initbuflen = ZIN0(1); //was seconds is now samples
	//Print("unit->m_delaytime %g\n", unit->m_delaytime);
	unit->m_fbufnum = -1.;
	
	unit->srate = SAMPLERATE;
    unit->srate_ms = unit->srate/1000.;
    //used bufferlength
    unit->initbuflen = unit->m_delaytime;
    unit->buflen = unit->initbuflen;
    unit->maxsize = unit->buflen / 3.;
    unit->onethirdBufsize = unit->maxsize;
    unit->minsize = MINSIZE;
	unit->semitone = sc_pow(2., 1./12.);
    
    unit->time = 0;	
   
    unit->gimme = 0.f;	
    unit->maxvoices = 100;
	//unit->doHanning = 0;
	
	int size1 = NUMVOICES * sizeof(voiceInfo);
	int size2 = PITCHTABLESIZE * sizeof(double);
	
	unit->voices = (voiceInfo*)RTAlloc(unit->mWorld, size1);
	unit->pitchTable = (double*)RTAlloc(unit->mWorld, size2);
	/*
	char* space = (char*)RTAlloc(unit->mWorld, size1 + size2);
	unit->voices = (voiceInfo*)space;
	space += size1;
	unit->pitchTable = (double*)space;
	*/
    for(i=0; i<NUMVOICES; i++) {
    	(unit->voices + i)->gvoiceSize= 100;			
    	(unit->voices + i)->gvoiceSpeed = 1.;			
    	(unit->voices + i)->gvoiceCurrent = 0.;
    	(unit->voices + i)->gvoiceDirection = 1;		
    	(unit->voices + i)->gvoiceOn = 0;			
    	(unit->voices + i)->gvoiceDone = 0;			
    	(unit->voices + i)->gvoiceRPan = .5;
    	(unit->voices + i)->gvoiceLPan = .5;
    	(unit->voices + i)->gvoiceGain = 1.;
		(unit->voices + i)->gvoiceOneOverRamp = 0.;
		(unit->voices + i)->gvoiceRamp = 0.;
    }
    //tempered scale
    
   //for(i=0; i<PITCHTABLESIZE-1; i += 2) {
   for(i=0; i<99; i += 2) {
		unit->pitchTable[i] = (double) 0.5*i;
		unit->pitchTable[i+1] =  (double)  -0.5*i;
	}
    unit->rampLength = 512.;
	//Munger_next(unit,1);
}




void MungerMono_next(MungerMono *unit, int inNumSamples)
{
	
	float grate, grate_var, glen, glen_var, ggain, ggain_var, position, delaytime, rampLength;
	//float grateRes = 0.f;
	int32 ambi;
	//int recordOn;
	float outsampL, outsampR;
	int newvoice;
	
	float *out = 	ZOUT(0);
	//float *outR = 	ZOUT(1);  
	delaytime = 	ZIN0(1); //is fixed not needed 
	grate 		= 	ZIN0(2);
	grate_var 	= 	ZIN0(3);
	glen 		= 	ZIN0(4);
	glen_var 	= 	ZIN0(5);
	double gpitch = (double) ZIN0(6);
	double gpitch_var = (double) ZIN0(7);
	ggain		= 	ZIN0(8); 
	ggain_var	= 	ZIN0(9); 
		
	ambi = (int32)	ZIN0(10);
	position = 		ZIN0(11);
	int32 numVoices = (int32) ZIN0(12);
	//gpan_var	= 	ZIN0(13); 
	int32 smoothPitch = (int32) ZIN0(13);
	rampLength = 	ZIN0(14);
	//unit->rampLength = rampLength;
	//unit->gpan_spread = gpan_var;
	
	GET_BUF
	CHECK_BUF
	
	voiceInfo * voiceinfos =  unit->voices;
	
	float gimme = unit->gimme;
	int32 time = unit->time;
	//other munger:
	long buflen = (long) unit->buflen;
	//float dsamp = unit->buflen;
	double semitone = unit->semitone;
	
	float newsize;
	int32 pitchChoice;
	double pitchExponent;
	
	double alpha, om_alpha, output, where;
	long first;
	long done, tail;
	
	
	
	double newPosition;	
	float value = 0.f;
	
	int foundOne;
	newvoice = 0;
	outsampR = 0.f;
	outsampL = 0.f;
	
	double * pitchTable = unit->pitchTable;

	float srate_ms = unit->srate_ms;
	float onethirdBufsize = unit->onethirdBufsize;
	float maxsize = unit->maxsize;
	float minsize = unit->minsize;
	RGET
/*
	float* dlybuf1 = bufData - ZOFF;
	float* dlywr   = dlybuf1 + (iwrphase & mask);
	float* dlyN    = dlybuf1 + bufSamples;	
*/
	
for (int32 k=0; k<inNumSamples; ++k) { 

///--------------
		if(time++ >= (long)gimme) 
		{
			time = 0;
///--------------- find voice:
			int i = 0; 
			foundOne = -1;
			while(foundOne < 0 && i < numVoices) { //was voices
				if (!(voiceinfos +i)->gvoiceOn) foundOne = i;
				i++;
			}
			newvoice = foundOne;
//-------
		if(newvoice >= 0) {
	
////// new Size---------------
	
		//int fpitch = sc_pow((float) unit->semitone, (float) pitchExponent); //something wrong with the pitch!@??
	//set grain pitch
			if(smoothPitch == 1) gpitch = gpitch +  ( (double) frand2(s1,s2,s3)) * gpitch_var;
			else {
				//temp = (float) frand(s1,s2,s3) * gpitch_var * 10.f; //100 * 0.1f;
				//pitchChoice = (int32) temp;
				
				pitchChoice = (int32) ( ((double) frand(s1,s2,s3) ) * gpitch_var * 10.0);
				pitchChoice = (int32) sc_clip(pitchChoice, 0, 99);
				pitchExponent = pitchTable[pitchChoice];
				gpitch =  gpitch * sc_pow(semitone, pitchExponent);
				
			}
			(voiceinfos +newvoice)->gvoiceSpeed =  gpitch;
			//if(g_voiceSpeed[newvoice] < (double) MINSPEED) .001 
			//if(gpitch < 0.001) g_voiceSpeed[newvoice] = (double) 0.001;
			//g_voiceSpeed[newvoice] = (double) sc_clip(gpitch, 0.001, 400.0);
			
			newsize = srate_ms * (glen + frand2(s1,s2,s3) * glen_var);
			
			if(newsize > maxsize) newsize = maxsize;
			if(newsize * gpitch > maxsize) newsize = maxsize/gpitch;
			if(newsize < minsize) newsize = minsize;
			(voiceinfos +newvoice)->gvoiceSize =  (long) newsize;
	//end new size
	
//newDirection------------------
			if(ambi == 0) {
				(voiceinfos +newvoice)->gvoiceDirection = (int) frand2(s1,s2,s3);
				if((voiceinfos +newvoice)->gvoiceDirection < 0) 
					(voiceinfos +newvoice)->gvoiceDirection = -1;
				else (voiceinfos +newvoice)->gvoiceDirection = 1;
			} else {
				if(ambi < 0) (voiceinfos +newvoice)->gvoiceDirection= -1; //was ambi==-1
				else (voiceinfos +newvoice)->gvoiceDirection = 1;
			}
//end newDirection-----------------
			//float nPan = frand2(s1,s2,s3) * gpan_var + 0.5f;
			//float rPan =  1.f - nPan;
			//(voiceinfos +newvoice)->gvoiceLPan 	= nPan;
			//(voiceinfos +newvoice)->gvoiceRPan	= rPan;
			(voiceinfos +newvoice)->gvoiceOn 	= 1;
			(voiceinfos +newvoice)->gvoiceDone	= 0;
			(voiceinfos +newvoice)->gvoiceGain = ggain + frand2(s1,s2,s3) * ggain_var;
//envelope
		float gvoiceRamp = (voiceinfos +newvoice)->gvoiceRamp;
			if(((voiceinfos +newvoice)->gvoiceSize) < 2.*rampLength) {
				gvoiceRamp = .5 * (voiceinfos + newvoice)->gvoiceSize;
				if(gvoiceRamp <= 0.) gvoiceRamp = 1.;
			(voiceinfos +newvoice)->gvoiceOneOverRamp = 1./gvoiceRamp;
			} else {
				gvoiceRamp = rampLength;
				if(gvoiceRamp <= 0.) gvoiceRamp = 1.;
				(voiceinfos +newvoice)->gvoiceOneOverRamp = 1./gvoiceRamp;
			}
			
	 (voiceinfos +newvoice)->gvoiceRamp = gvoiceRamp;
/*** set start point; tricky, cause of moving buffer, variable playback rates, backwards/forwards, etc.... ***/
	
	// 3. random positioning and fixed buffer	
			//else if (position == -1. && recordOn == 0) {
			 if (position == -1.) {
				if((voiceinfos +newvoice)->gvoiceDirection == 1) {//going forward			
					newPosition = (double) onethirdBufsize - frand(s1,s2,s3) * onethirdBufsize;
				}	
				else //going backwards
					newPosition = (double) frand(s1,s2,s3) * onethirdBufsize;
				}
	
	// 4. fixed positioning and fixed buffer	
		//	else if (position >= 0. && recordOn == 0) {
			else if (position >= 0.) {
				if((voiceinfos +newvoice)->gvoiceDirection == 1) {//going forward			
					newPosition = (double) onethirdBufsize - position * onethirdBufsize;
				}	
			else //going backwards
				newPosition = (double) position * onethirdBufsize;
			}
			
			(voiceinfos +newvoice)->gvoiceCurrent = newPosition;
///////end new setup---------------------------------------------------------
		}
		//grateRes = grate  + frand(s1,s2,s3) * grate_var;
						//gimme = unit->srate_ms * grate; //grate is actually time-distance between grains
		gimme = srate_ms * (grate  + frand(s1,s2,s3) * grate_var);
		}
		
//mix 'em, pan 'em -------------------------------------
	//i = voices;
	//while(i--){
	for(int j = 0; j<numVoices; j++){
		if((voiceinfos + j)->gvoiceOn) {
			where = (voiceinfos + j)->gvoiceCurrent;

//getSamp -/lin interpol------------------------------------------------
			while(where < 0.) where += buflen;
			while(where > guardFrame) where -= buflen;

			first = (long)where;
			alpha = (double) where - first;
			om_alpha = 1. - alpha;		
			output = bufData[first++ & mask] * om_alpha;
			if(first <  buflen) {
				output += bufData[first & mask] * alpha;
			}
			else {
				output +=  bufData[0 & mask] * alpha;
			}
			outsampL = (float) output;
	
	
//envelope---------------------------------------------------

			done = (voiceinfos + j)->gvoiceDone;
			tail = (voiceinfos + j)->gvoiceSize - done;
	
			if(done < (voiceinfos + j)->gvoiceRamp) outsampL *= (done * (voiceinfos + j)->gvoiceOneOverRamp);
			else if(tail < (voiceinfos + j)->gvoiceRamp) outsampL *= (tail * (voiceinfos + j)->gvoiceOneOverRamp);
	
//end env---------------------
			outsampL = outsampL * (voiceinfos + j)->gvoiceGain;
		//pan it, dumb linear
			value += outsampL;
			//outsampR += outsampL * (voiceinfos + j)->gvoiceRPan;
		//see if grain is done after jumping to next sample point
			(voiceinfos + j)->gvoiceCurrent += (double) (voiceinfos + j)->gvoiceDirection * (voiceinfos + j)->gvoiceSpeed;
			if(++(voiceinfos + j)->gvoiceDone >= (voiceinfos + j)->gvoiceSize) (voiceinfos + j)->gvoiceOn = 0;
				}
			}//end mix em			
				ZXP(out) = value;
				//ZXP(outR) = outsampR;

	value = 0.f;
	//just make sure to reset ereything
	newvoice = 0;
	outsampR = 0.f;
	outsampL = 0.f;
		}
	RPUT
	//write back :	
	unit->gimme = gimme;
	unit->time = time;
	unit->voices = voiceinfos;

}
				
/////////////////////RecPBufDelayL////////////////////////

void RecPBufDelayL_Ctor(RecPBufDelayL *unit)
{
	if (INRATE(1) == calc_FullRate) {
		SETCALC(RecPBufDelayL_next_z);
	} else {
		//SETCALC(RecPBufDelayL_next_ak);
	}
	unit->m_delaytime = ZIN0(2);
	//Print("unit->m_delaytime %g\n", unit->m_delaytime);
	//unit->m_dlybuf = 0;
	unit->m_fbufnum = -1.;
	
	//DelayUnit_AllocDelayLine(unit);
	//Print("->GET_BUF\n");
	GET_BUF
	//Print("<-GET_BUF\n");
	unit->m_dsamp = BufCalcDelay(unit->m_delaytime);
	unit->m_numoutput = 0;
	unit->m_iwrphase = 0;
	unit->m_x1 = 0.f;
	unit->m_y1 = 0.f;
	unit->m_recordOn = (long) ZIN0(3);
	
	ZOUT0(0) = 0.f;
}



///////////////////////////////////////////////////////////////////////////
void RecPBufDelayL_next(RecPBufDelayL *unit, int inNumSamples)
{	
	float *out = ZOUT(0);
	float *in = ZIN(1);
	float delaytime = ZIN0(2);
	long recordOn = (long) ZIN0(3);
	
	
	GET_BUF
	long iwrphase = unit->m_iwrphase;
	float dsamp = unit->m_dsamp;

	if (delaytime == unit->m_delaytime) {
		long irdphase = iwrphase - (long)dsamp;
		float* dlybuf1 = bufData - ZOFF;
		float* dlyrd   = dlybuf1 + (irdphase & mask);
		float* dlywr   = dlybuf1 + (iwrphase & mask);
		float* dlyN    = dlybuf1 + bufSamples;
		long remain = inNumSamples;
		while (remain) {
			long rdspace = dlyN - dlyrd;
			long wrspace = dlyN - dlywr;
			long nsmps = sc_min(rdspace, wrspace);
			if (nsmps == 0)return; // buffer not allocated.
			nsmps = sc_min(remain, nsmps);
			remain -= nsmps;
			LOOP(nsmps,
				float value = ZXP(dlyrd);
				if(recordOn)
				ZXP(dlywr) = ZXP(in);
				ZXP(out) = value;
			);
			if (dlyrd == dlyN) dlyrd = dlybuf1;
			if (dlywr == dlyN) dlywr = dlybuf1;
		}
		iwrphase += inNumSamples;
	} else {
		float next_dsamp = BufCalcDelay(delaytime);
		float dsamp_slope = CALCSLOPE(next_dsamp, dsamp);
		
		LOOP(inNumSamples,
			dsamp += dsamp_slope;
			++iwrphase;
			long irdphase = iwrphase - (long)dsamp;
			float value = bufData[irdphase & mask];
			if(recordOn)
			bufData[iwrphase & mask] = ZXP(in);
			ZXP(out) = value;
		);
		unit->m_dsamp = dsamp;
		unit->m_delaytime = delaytime;
	}
	
	unit->m_iwrphase = iwrphase;
	unit->m_recordOn = recordOn;

}
void RecPBufDelayL_next_z(RecPBufDelayL *unit, int inNumSamples)
{	
	float *out = ZOUT(0);
	float *in = ZIN(1);
	float delaytime = ZIN0(2);
	long recordOn = (long) ZIN0(3);
	
	GET_BUF
	long iwrphase = unit->m_iwrphase;
	float dsamp = unit->m_dsamp;

	if (delaytime == unit->m_delaytime) {
		long irdphase = iwrphase - (long)dsamp;
		float* dlybuf1 = bufData - ZOFF;
		float* dlyN    = dlybuf1 + bufSamples;
		long remain = inNumSamples;
		while (remain) {
			float* dlywr = dlybuf1 + (iwrphase & mask);
			float* dlyrd = dlybuf1 + (irdphase & mask);
			long rdspace = dlyN - dlyrd;
			long wrspace = dlyN - dlywr;
			long nsmps = sc_min(rdspace, wrspace);
			if (nsmps == 0) return; // buffer not allocated.
			nsmps = sc_min(remain, nsmps);
			remain -= nsmps;
			if (irdphase < 0) {
				LOOP(nsmps,
					if(recordOn)
					ZXP(dlywr) = ZXP(in);
					ZXP(out) = 0.f;
				);
			} else {
				LOOP(nsmps,
					float value = ZXP(dlyrd);
					if(recordOn)
					ZXP(dlywr) = ZXP(in);
					ZXP(out) = value;
				);
			}
			iwrphase += nsmps;
			irdphase += nsmps;
		}
	} else {
		
		float next_dsamp = BufCalcDelay(delaytime);
		float dsamp_slope = CALCSLOPE(next_dsamp, dsamp);
		
		LOOP(inNumSamples,
			dsamp += dsamp_slope;
			long irdphase = iwrphase - (long)dsamp;
			
			if (irdphase < 0) {
				if(recordOn)
				bufData[iwrphase & mask] = ZXP(in);
				ZXP(out) = 0.f;
			} else {
				float value = bufData[irdphase & mask];
				if(recordOn)
				bufData[iwrphase & mask] = ZXP(in);
				ZXP(out) = value;
			}
			iwrphase++;
		);
		unit->m_dsamp = dsamp;
		unit->m_delaytime = delaytime;
	}
	
	unit->m_iwrphase = iwrphase;
	unit->m_recordOn = recordOn;
	unit->m_numoutput += inNumSamples;
	if (unit->m_numoutput >= bufSamples) {
		SETCALC(RecPBufDelayL_next);
	}
}



void ImpulseDrop_next_a(ImpulseDrop *unit, int inNumSamples)
{
	float *out = ZOUT(0);
	float *freq = ZIN(0);
	float *prob = ZIN(2);
        float randNum;
        RGET
	float freqmul = unit->mFreqMul;
	double phase = unit->mPhase;

	LOOP(inNumSamples, 
		float z;
                                
		if (phase >= 1.f) {
                    randNum = frand(s1,s2,s3);
                  //  Print("rand is %f \n", randNum);

                    if(ZXP(prob) < randNum)
			z = 1.f;
                    else
                        z=0.f;
                        
                    phase -= 1.f;

		} else {
			z = 0.f;
		}
		phase += ZXP(freq) * freqmul;
		ZXP(out) = z;
                
	);

	unit->mPhase = phase;
	RPUT
}

void ImpulseDrop_next_k(ImpulseDrop *unit, int inNumSamples)
{
	float *out = ZOUT(0);
	float freq = ZIN0(0) * unit->mFreqMul;
	float prob = ZIN0(2);
        float randNum;
        RGET
     //   Print("in krate");

	double phase = unit->mPhase;
	LOOP(inNumSamples, 
		float z;
		if (phase >= 1.f) {
                    randNum = frand(s1,s2,s3);
                   // Print("rand is %f \n", randNum);

                    if(prob < randNum)
			z = 1.f;
                    else
                        z=0.f;
                        
                    phase -= 1.f;
		} else {
			z = 0.f;
		}
		phase += freq;
		ZXP(out) = z;
	);

	unit->mPhase = phase;
        RPUT
}

void ImpulseDrop_Ctor(ImpulseDrop* unit)
{
	if (INRATE(0) == calc_FullRate) {
		SETCALC(ImpulseDrop_next_a);
	} else {
		SETCALC(ImpulseDrop_next_k);
	}

	unit->mFreqMul = unit->mRate->mSampleDur;
	unit->mPhase = ZIN0(1);
        unit->mProb = ZIN0(2); 
       // Print("prob is %f \n", unit->mProb);


	if (unit->mPhase == 0.f) unit->mPhase = 1.f;
	ZOUT0(0) = 0.f;

}
//////////////////////////////////////////////////////////////////////////////////////////
void PulseCount2_Ctor(PulseCount2 *unit)
{
	SETCALC(PulseCount2_next);

	unit->m_prevtrig = 0.f;
	unit->m_prevreset = 0.f;
	unit->mLevel = 0.f;
	
	ZOUT0(0) = 0.f;
}


void PulseCount2_next(PulseCount2 *unit, int inNumSamples)
{
	float *out = ZOUT(0);
	float *trig = ZIN(0);
	float *reset = ZIN(1);
	float prevtrig = unit->m_prevtrig;
	float prevreset = unit->m_prevreset;
	float level = unit->mLevel;

	LOOP(inNumSamples, 
		float curtrig = ZXP(trig);
		float curreset = ZXP(reset);
		if (prevreset <= 0.f && curreset > 0.f && level > 0.f) level -= 1.f;
		else if (prevtrig <= 0.f && curtrig > 0.f) level += 1.f;
		ZXP(out) = level;
		prevtrig = curtrig;
		prevreset = curreset;
	);
	unit->mLevel = level;
	unit->m_prevtrig = prevtrig;
	unit->m_prevreset = prevreset;
}
//////////////////////////////////////////////////////////////////////////////////////////
void PulseCount3_Ctor(PulseCount3 *unit)
{
	SETCALC(PulseCount2_next);

	unit->m_prevtrig = 0.f;
	unit->m_prevreset = 0.f;
	unit->mLevel = 0.f;
	
	ZOUT0(0) = 0.f;
}


void PulseCount3_next(PulseCount3 *unit, int inNumSamples)
{
	float *out = ZOUT(0);
	float *trig = ZIN(0);
	float *max = ZIN(1);
	float prevtrig = unit->m_prevtrig;
	float prevreset = unit->m_prevreset;
	float level = unit->mLevel;

	LOOP(inNumSamples, 
		float curtrig = ZXP(trig);
		float curmax = ZXP(max);
		if (curmax > 0.f && level >= curmax) level = 0.f;
		else if (prevtrig <= 0.f && curtrig > 0.f) level += 1.f;
		ZXP(out) = level;
		prevtrig = curtrig;
		//prevreset = curreset;
	);
	unit->mLevel = level;
	unit->m_prevtrig = prevtrig;
	//unit->m_prevreset = prevreset;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
void TIUrn_Ctor(TIUrn* unit)
{	
	int lo = (int)ZIN0(0);
	int hi = (int)ZIN0(1);
	int range = hi - lo + 1;
	//unit->lastValues = (float*)RTAlloc(unit->mWorld, range * sizeof(float));
	unit->index = 0;
	RGen& rgen = *unit->mParent->mRGen;	
	float outval = (float)(rgen.irand(range) + lo);
	ZOUT0(0) = outval;
	unit->lastValue = outval;
	SETCALC(TIUrn_next);
	unit->m_trig = ZIN0(2);
	///
	//setsysbar();

}

void TIUrn_next(TIUrn* unit, int inNumSamples)
{	
	float trig = ZIN0(2);
	float lastVal = unit->lastValue;
	if (trig > 0.f && unit->m_trig <= 0.f) {
		float outval;
		int lo = (int)ZIN0(0);
		int hi = (int)ZIN0(1);
		int range = hi - lo + 1;
		RGen& rgen = *unit->mParent->mRGen;	
		
		outval = (float)(rgen.irand(range) + lo);
		while(outval == lastVal) outval = (float)(rgen.irand(range) + lo);
		
		ZOUT0(0) = unit->m_value = (float)outval;
		unit->lastValue = (float) outval;
	} else {
		ZOUT0(0) = unit->m_value;
	}
	unit->m_trig = trig;
}

void TIUrn_Dtor(TIUrn* unit)
{
	//RTFree(unit->mWorld, unit->lastValues); 
}
///////////////////////////////////////////////////

void load(InterfaceTable *inTable)
{
	ft = inTable;

	DefineSimpleUnit(RecPBufDelayL);
	DefineDtorUnit(Munger);
	DefineDtorUnit(MungerMono);
	DefineSimpleUnit(ImpulseDrop);
	DefineSimpleUnit(PulseCount2);
	DefineSimpleUnit(PulseCount3);
	DefineDtorUnit(TIUrn);

//        DefineSimpleUnit(AltivecUnit);
	
}

///////////////////////////////////////////////////////////////////////////
